
#include "triangulation.h"


#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#endif



static int error_indicator = TRIANGULATION_OK;
static int error_on_line   = 0;

typedef struct _triangulation_t {
  char                     *file;
  int                       nnodes;
  int                       ntriangles;
  dict_dictionary           noded;
  dict_dictionary           triangled;
  dict_dictionary           groupd;
  dict_dictionary           clientd;
  triangulation_client_t    curr_client;
  unsigned long int         curr_client_idx;
}              _triangulation_t;

typedef struct _triangulation_node_t {
  struct _triangulation_t                *t;
  int                                     id;
  float                                   x, y;
  triangulation_assoc_data_t             *assoc_data;
  LIST_OF (_triangulation_triangle_t *)   incid_tri_l;
}              _triangulation_node_t;


typedef struct _triangulation_group_t {
  int   id;
  char  *name;
}              _triangulation_group_t;

typedef struct _triangulation_triangle_t {
  struct _triangulation_t                *t;
  int                                     id;
  _triangulation_node_t                  *nodes[3];
  _triangulation_group_t                 *group;
  triangulation_assoc_data_t             *assoc_data;
}              _triangulation_triangle_t;


static _triangulation_group_t *
find_group (triangulation_t t, char *group)
{
  _triangulation_group_t *g;
  static int id = 0;

  g = dict_find_obj_with_string_name (t->groupd, group);
  if (g == NULL) {
    g = NEWP_OBJ (_triangulation_group_t);
    if (g == NULL) {
      error_indicator = TRIANGULATION_MALLOC; 
      return NULL;
    }
    g->id = id++;
    NEWP_DUPL_STRING (g->name, group,
                      {
                        error_indicator = TRIANGULATION_MALLOC;
                        return NULL;
                      });
    dict_add_obj_string_name (t->groupd, g, group);
  }
  return g;
}

static int
num_clients (_triangulation_t *t)
{
  return dict_count (t->clientd);
}

static int
client_num (triangulation_t t, triangulation_client_t client)
{
  int cn = (int) dict_find_obj_with_handle_name (t->clientd, client);
  return cn;
}

static int
have_client (triangulation_t t, triangulation_client_t client)
{
  return (client_num (t, client));
}

static triangulation_assoc_data_t
realloc_assoc_data (triangulation_assoc_data_t ad, int newlen)
{
  triangulation_assoc_data_t *newad = NULL;

  if (newlen <= 0) return NULL;
  
  if (ad == NULL) {
    int i;
    newad =   NEWP_OBJ_ARRAY (newlen, triangulation_assoc_data_t);
    for (i = 0; i < newlen; i++) { newad[i] = NULL; }
  } else {
    newad = NEWP_REALLOC_OBJ_ARRAY (ad, newlen, triangulation_assoc_data_t);
  }
  if (newad != NULL) {
    newad[newlen-1] = NULL;
  }
  return newad;
}

static void
node_attach_tri (_triangulation_node_t *n, _triangulation_triangle_t *tri)
{
  add_to_tail (n->incid_tri_l, tri);
}

triangulation_triangle_p 
tri_add_triangle (triangulation_t t,
                  int id,
                  triangulation_node_p n[3],
                  char *group
                  )
{
  _triangulation_triangle_t *tri;
  _triangulation_group_t *g = NULL;
  
  tri = NEWP_OBJ (_triangulation_triangle_t);
  if (tri == NULL) {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  /* attach to triangulation */
  tri->id = id;
  tri->t  = t;
  dict_add_obj_int_name (t->triangled, tri, tri->id);
  {
    int i;
    for (i = 0; i < 3; i++) {
      tri->nodes[i] = n[i];
      node_attach_tri (tri->nodes[i], tri);
    }
  }

  g = find_group (t, group);
  if (g == NULL)
    return NULL;
  tri->group = g;
  tri->assoc_data = realloc_assoc_data (NULL, num_clients (t));
  
  return tri;
}

static _triangulation_triangle_t *
read_triangle (triangulation_t t, FILE *stream, int i)
{
  int id, n0, n1, n2;
  int matid, soilid;
  float thickness;
  char group[32];
  _triangulation_node_t *n[3]={NULL,NULL,NULL};

  if (fscanf (stream, "%d %d %d %d %d %f %d %s",
              &id, &n0, &n1, &n2, &matid, &thickness, &soilid, group) != 8) {
    error_indicator = TRIANGULATION_BAD_TRIANGLE;
    error_on_line   = t->nnodes + i + 4 + 1;
    return NULL;
  }

  n[0] = dict_find_obj_with_int_name (t->noded, n0);
  if (n[0] == NULL) {
    error_indicator = TRIANGULATION_BAD_TRIANGLE;
    error_on_line   = t->nnodes + i + 4 + 1;
    return NULL;
  }

  n[1] = dict_find_obj_with_int_name (t->noded, n1);
  if (n[1] == NULL) {
    error_indicator = TRIANGULATION_BAD_TRIANGLE;
    error_on_line   = t->nnodes + i + 4 + 1;
    return NULL;
  }

  n[2] = dict_find_obj_with_int_name (t->noded, n2);
  if (n[2] == NULL) {
    error_indicator = TRIANGULATION_BAD_TRIANGLE;
    error_on_line   = t->nnodes + i + 4 + 1;
    return NULL;
  }
  
  return tri_add_triangle (t, id, n, group);
}

triangulation_node_p
tri_add_node (triangulation_t t, int id, double x, double y)
{
  _triangulation_node_t *n;
  n = NEWP_OBJ (_triangulation_node_t);
  if (n == NULL) {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  /* attach to triangulation */
  n->id          = id;
  n->t           = t;
  dict_add_obj_int_name (t->noded, n, n->id);
  n->x           = x;
  n->y           = y;
  n->assoc_data  = realloc_assoc_data (NULL, num_clients (t));
  n->incid_tri_l = make_list ();

  return n;
}

static _triangulation_node_t *
read_node (triangulation_t t, FILE *stream, int i)
{
  int id;
  float x, y;

  if (fscanf (stream, "%d %f %f", &id, &x, &y) != 3) {
    error_indicator = TRIANGULATION_BAD_NODE;
    error_on_line   = i + 3 + 1;
    return NULL;
  }

  return tri_add_node (t, id, x, y);
}

triangulation_t
tri_new_triangulation (void)
{
  triangulation_t t;
  
  error_indicator = TRIANGULATION_OK;
  error_on_line   = 0;
  
  t = NEWP_OBJ (_triangulation_t);
  if (t == NULL) {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }

  t->clientd = dict_create_dictionary ();
  if (!t->clientd) {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  t->curr_client     = NULL;
  t->curr_client_idx = 0;

  t->noded = dict_create_dictionary ();
  if (t->noded == NULL)  {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  t->triangled = dict_create_dictionary ();
  if (t->triangled == NULL)  {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  t->groupd = dict_create_dictionary ();
  if (t->groupd == NULL)  {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }
  
  NEWP_DUPL_STRING (t->file, "File name not set",
                    {error_indicator = TRIANGULATION_MALLOC; return NULL;}
                    );

  t->nnodes = 0;
  t->ntriangles = 0;

  return t;
}

triangulation_t
tri_from_targe2_file (FILE *stream)
{
  char buf[512];
  triangulation_t t;
  int i;
  _triangulation_node_t *n;
  _triangulation_triangle_t *tri;
  
  error_indicator = TRIANGULATION_OK;
  error_on_line   = 0;

  t = tri_new_triangulation ();
  if (t == NULL) {
    error_indicator = TRIANGULATION_MALLOC; 
    return NULL;
  }

  /* read file name */
  if (t->file) NEWP_FREE_OBJ (t->file);
  fscanf (stream, "%s", buf);
  NEWP_DUPL_STRING (t->file, buf,
                    {error_indicator = TRIANGULATION_MALLOC; return NULL;}
                    );

  /* read totals */
  if (fscanf (stream, "%d %d", &t->nnodes, &t->ntriangles) != 2) {
    error_indicator = TRIANGULATION_BAD_TOTALS;
    return NULL;
  }

  if (t->nnodes <= 0) {
    error_indicator = TRIANGULATION_BAD_TOTALS;
    return NULL;
  }
  if (t->ntriangles <= 0) {
    error_indicator = TRIANGULATION_BAD_TOTALS;
    return NULL;
  }

  /* skip *NODES */
  fscanf (stream, "%s", buf);

  /* read nodes */
  for (i = 0; i < t->nnodes; i++) {
    n = read_node (t, stream, i);
    if (n == NULL)
      return NULL;
  }
  
  /* skip *TR3N */
  fscanf (stream, "%s", buf);

  /* read triangles */
  for (i = 0; i < t->ntriangles; i++) {
    tri = read_triangle (t, stream, i);
    if (tri == NULL)
      return NULL;
  }

  return t;
}

triangulation_t
tri_from_afro_file (FILE *stream)
{
  return tri_from_targe2_file (stream);
}

triangulation_status_t
tri_status (void)
{
  return (triangulation_status_t) error_indicator;
}

int
tri_error_on_line (void)
{
  return error_on_line;
}

/*
  Iterate nodes.
 */
triangulation_node_p
tri_next_node (triangulation_t t, dict_iterator iterator)
{
  return (triangulation_node_p) dict_iter_objs (t->noded, iterator);
}

/*
  Iterate triangles.
 */
triangulation_triangle_p
tri_next_triangle (triangulation_t t, dict_iterator iterator)
{
  return (triangulation_triangle_p) dict_iter_objs (t->triangled, iterator);
}

/*
  Find node by id.
 */
triangulation_node_p
tri_find_node (triangulation_t t, int node_id)
{
  triangulation_node_p n = (triangulation_node_p) dict_find_obj_with_int_name (t->noded, node_id);
  return n;
}

/*
  Find triangle by id.
 */
triangulation_triangle_p
tri_find_triangle (triangulation_t t, int id)
{
  return (triangulation_triangle_p) dict_find_obj_with_int_name (t->triangled, id);
}

static int
realloc_assoc_data_fields (triangulation_t t)
{
  triangulation_node_p n;
  triangulation_triangle_p tri;
  int newlen = num_clients (t);
  dict_iterator di = dict_get_iterator ();

  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    n->assoc_data = realloc_assoc_data (n->assoc_data, newlen);
    if (!n->assoc_data) {
      error_indicator = TRIANGULATION_MALLOC; 
      return FALSE;
    }
  }

  dict_reset_iterator (di);
  while (tri = tri_next_triangle (t, di)) {
    tri->assoc_data = realloc_assoc_data (tri->assoc_data, newlen);
    if (!tri->assoc_data) {
      error_indicator = TRIANGULATION_MALLOC; 
      return FALSE;
    }
  }

  dict_del_iterator (di);
  return TRUE;
}

static int
set_up_for_client (triangulation_t t, triangulation_client_t client)
{
  if (t->curr_client != client) {
    int cn = client_num (t, client);
    if (cn == 0) { /* don't have this client yet */
      cn = num_clients (t) + 1;
      dict_add_obj_handle_name (t->clientd, (dict_object) cn, client);
      if (!realloc_assoc_data_fields (t)) {
        return FALSE;
      }
    }
    t->curr_client     = client;
    t->curr_client_idx = cn-1;
  }
  return TRUE;
}

/*
  Get associated data.
 */
triangulation_assoc_data_t
tri_node_get_assoc_data (triangulation_node_p n, triangulation_client_t client)
{
  if (client == NULL) return NULL;
  if (have_client (n->t, client)) {
    if (set_up_for_client (n->t, client)) {
      return n->assoc_data[n->t->curr_client_idx];
    }
  }
  return NULL;
}

triangulation_assoc_data_t
tri_node_set_assoc_data (triangulation_node_p n, triangulation_client_t client,
                         triangulation_assoc_data_t newad)
{
  if (client == NULL) return NULL;
  if (set_up_for_client (n->t, client)) {
    triangulation_assoc_data_t oldad = n->assoc_data[n->t->curr_client_idx];
    n->assoc_data[n->t->curr_client_idx] = newad;
    return oldad;
  }
  return NULL;
}

/*
  Get associated data.
 */
triangulation_assoc_data_t
tri_triangle_get_assoc_data (triangulation_triangle_p tri, triangulation_client_t client)
{
  if (client == NULL) return NULL;
  if (have_client (tri->t, client)) {
    if (set_up_for_client (tri->t, client)) {
      return tri->assoc_data[tri->t->curr_client_idx];
    }
  }
  return NULL;
}

triangulation_assoc_data_t
tri_triangle_set_assoc_data (triangulation_triangle_p tri, triangulation_client_t client,
                         triangulation_assoc_data_t newad)
{
  if (client == NULL) return NULL;
  if (set_up_for_client (tri->t, client)) {
    triangulation_assoc_data_t oldad = tri->assoc_data[tri->t->curr_client_idx];
    tri->assoc_data[tri->t->curr_client_idx] = newad;
    return oldad;
  }
  return NULL;
}

/*
  Compute average edge length.
 */
double
tri_avg_edge_length (triangulation_t t)
{
  _triangulation_triangle_t *tri;
  double avl = 0;
  double tottri;
  double xc[3], yc[3];
  double dx, dy;
  dict_iterator iterator = NULL;
  int i;
  
  tri = NULL;
  tottri = 0;
  iterator = dict_get_iterator ();
  dict_reset_iterator (iterator);
  
  while (tri = dict_iter_objs (t->triangled, iterator)) {
    for (i = 0; i < 3; i++) {
      xc[i] = tri->nodes[i]->x;
      yc[i] = tri->nodes[i]->y;
    }
    dx = xc[0] - xc[1]; dy = yc[0] - yc[1];
    avl += sqrt (dx*dx + dy*dy);
    dx = xc[2] - xc[1]; dy = yc[2] - yc[1];
    avl += sqrt (dx*dx + dy*dy);
    dx = xc[0] - xc[2]; dy = yc[0] - yc[2];
    avl += sqrt (dx*dx + dy*dy);
    tottri++;
  }
  dict_del_iterator (iterator);
  return avl / (2 * tottri);
}

/*
  Compute longest edge length.
 */
double
tri_max_edge_length (triangulation_t t)
{
  _triangulation_triangle_t *tri;
  double mel = 0, el;
  double xc[3], yc[3];
  double dx, dy;
  dict_iterator iterator = NULL;
  int i;
  
  tri = NULL;
  iterator = dict_get_iterator ();
  dict_reset_iterator (iterator);
  
  while (tri = dict_iter_objs (t->triangled, iterator)) {
    for (i = 0; i < 3; i++) {
      xc[i] = tri->nodes[i]->x;
      yc[i] = tri->nodes[i]->y;
    }
    dx  = xc[0] - xc[1]; dy = yc[0] - yc[1];
    el  = sqrt (dx*dx + dy*dy);
    mel = max (mel, el);
    dx  = xc[2] - xc[1]; dy = yc[2] - yc[1];
    el  = sqrt (dx*dx + dy*dy);
    mel = max (mel, el);
    dx  = xc[0] - xc[2]; dy = yc[0] - yc[2];
    el  = sqrt (dx*dx + dy*dy);
    mel = max (mel, el);
  }
  dict_del_iterator (iterator);
  return mel;
}

/*
  Compute longest length of edge incident on node.
 */
double
tri_max_edge_length_at_node (triangulation_node_p n)
{
  double mel, el;
  double xc[3], yc[3];
  double dx, dy;
  _triangulation_triangle_t *tri;
  int i;
  
  mel = 0;
  loop_through_list (n->incid_tri_l, tri, _triangulation_triangle_t *) {
    for (i = 0; i < 3; i++) {
      xc[i] = tri->nodes[i]->x;
      yc[i] = tri->nodes[i]->y;
    }
    if (tri->nodes[0] == n || tri->nodes[1] == n) {
      dx  = xc[0] - xc[1]; dy = yc[0] - yc[1];
      el  = sqrt (dx*dx + dy*dy);
      mel = max (mel, el);
    }
    if (tri->nodes[2] == n || tri->nodes[1] == n) {
      dx  = xc[2] - xc[1]; dy = yc[2] - yc[1];
      el  = sqrt (dx*dx + dy*dy);
      mel = max (mel, el);
    }
    if (tri->nodes[0] == n || tri->nodes[2] == n) {
      dx  = xc[0] - xc[2]; dy = yc[0] - yc[2];
      el  = sqrt (dx*dx + dy*dy);
      mel = max (mel, el);
    }
  }
  return mel;
}

int
tri_triangle_get_nodes (triangulation_triangle_p tri, triangulation_node_p n[3])
{
  int i;
  for (i = 0; i < 3; i++) n[i] = tri->nodes[i];
  return 1;
}

int
tri_triangle_get_id (triangulation_triangle_p tri)
{
  return tri->id;
}

int
tri_triangle_get_group (triangulation_triangle_p tri,                
                        int group_name_length, char *group,
                        int *group_id)
{
  int i, len;
  len    = strlen (tri->group->name);
  if (len > group_name_length) len = group_name_length;
  for (i = 0; i < len; i++) {
    group[i] = tri->group->name[i];
  }
  *group_id = tri->group->id;
  return TRUE;
}

int
tri_node_get_xy (triangulation_node_p n, double *x, double *y)
{
  *x = n->x;
  *y = n->y;
  return 1;
}

int
tri_node_set_xy (triangulation_node_p n, double x, double y)
{
  n->x = x;
  n->y = y;
  return 1;
}

triangulation_triangle_p
tri_node_next_incid_triangle (triangulation_node_p n, triangulation_triangle_p tri)
{
  return (triangulation_triangle_p) get_list_next (n->incid_tri_l, tri);
}

int
tri_node_num_incid_triangles (triangulation_node_p n)
{
  return list_length (n->incid_tri_l);
}

int
tri_node_get_id (triangulation_node_p n)
{
  return n->id;
}

static dict_dictionary
one_ring (triangulation_node_p n)
{
  triangulation_triangle_p tri;
  triangulation_node_p na[3];
  int i;
  dict_dictionary d = dict_create_dictionary ();
  if (!d) return NULL;
  tri = NULL;
  while (tri = tri_node_next_incid_triangle (n, tri)) {
    tri_triangle_get_nodes (tri, na);
    for (i = 0; i < 3; i++) {
      if (na[i] != n) {         /* not the center node */
        if (dict_find_obj_with_handle_name (d, na[i]) == NULL) { /* not in ring yet ... */
          dict_add_obj_handle_name (d, na[i], na[i]); /* ... add it */
        }
      }
    }
  }
  return d;
}

static dict_dictionary
two_ring (triangulation_node_p n)
{
  triangulation_triangle_p tri;
  triangulation_node_p an, na[3];
  int i;
  dict_iterator di;
  dict_dictionary d = dict_create_dictionary ();
  dict_dictionary ord = one_ring (n);
  if (!ord) return NULL;
  if (!d)   return NULL;
  
  di = dict_get_iterator ();
  dict_reset_iterator (di);
  while (an = dict_iter_objs (ord, di)) { /* for all nodes in the one ring */
    tri = NULL;
    while (tri = tri_node_next_incid_triangle (an, tri)) { /* for all tri incident on an */
      tri_triangle_get_nodes (tri, na);
      for (i = 0; i < 3; i++) {
        if (na[i] != an && na[i] != n) {
          triangulation_node_p in_ord = dict_find_obj_with_handle_name (ord, na[i]);
          triangulation_node_p in_d   = dict_find_obj_with_handle_name (d, na[i]);
          if (in_ord == NULL && /* not in one ring ... */
              in_d   == NULL /* ... not in the two ring yet ... */
              ) {
            dict_add_obj_handle_name (d, na[i], na[i]); /* ... add it */
          }
        }
      }
    }
  }
  dict_del_iterator (di);
  dict_destroy_dictionary (ord);
  return d;
}

dict_dictionary
tri_node_ring_nodes (triangulation_node_p n, int ring_num)
{
  switch (ring_num) {
  case 1:
    return one_ring (n);
  case 2:
    return two_ring (n);
  default:
    return NULL;                /* don't know how to handle */
  }
}

static void
destroy_node (triangulation_node_p n)
{
  if (n->assoc_data) NEWP_FREE_OBJ (n->assoc_data);
  if (n->incid_tri_l) free_list (n->incid_tri_l, NOT_NODES);
  NEWP_FREE_OBJ (n);
}

static void
destroy_triangle (triangulation_triangle_p tri)
{
  if (tri->assoc_data) NEWP_FREE_OBJ (tri->assoc_data);
  NEWP_FREE_OBJ (tri);
}

void
tri_destroy_triangulation (triangulation_t t)
{
  if (t->noded) {
    triangulation_node_p n;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (n = tri_next_node (t, di)) {
      destroy_node (n);
    }
    dict_del_iterator (di);
  }

  if (t->triangled) {
    triangulation_triangle_p tri;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (tri = tri_next_triangle (t, di)) {
      destroy_triangle (tri);
    }
    dict_del_iterator (di);
  }

  if (t->groupd) {
    _triangulation_group_t *group;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (group = dict_iter_objs (t->groupd, di)) {
      if (group->name) NEWP_FREE_OBJ (group->name);
      NEWP_FREE_OBJ (group);
    }
    dict_del_iterator (di);
  }
  
  if (t->file) NEWP_FREE_OBJ (t->file);
  if (t->noded) dict_destroy_dictionary (t->noded);
  if (t->triangled) dict_destroy_dictionary (t->triangled);
  if (t->groupd) dict_destroy_dictionary (t->groupd);
  if (t->clientd) dict_destroy_dictionary (t->clientd);
  NEWP_FREE_OBJ (t);
}

static double
area (double x1, double y1, double x2, double y2, double x3, double y3) 
{
    double a1, a2, a3;

    a1 = x2 * y3 - x3 * y2;
    a2 = x3 * y1 - x1 * y3;
    a3 = x1 * y2 - x2 * y1;
    return 0.5 * (a1 + a2 + a3);
} 

double
tri_triangle_get_area (triangulation_triangle_p tri)
{
  return area (tri->nodes[0]->x, tri->nodes[0]->y,
               tri->nodes[1]->x, tri->nodes[1]->y,
               tri->nodes[2]->x, tri->nodes[2]->y);
}

int
tri_to_targe2_file (triangulation_t t, FILE *stream)
{
  triangulation_node_p n;
  triangulation_triangle_p tri;
  dict_iterator di = dict_get_iterator ();
  fprintf (stream, "%s\n", t->file);
  fprintf (stream, "%d %d\n", dict_count (t->noded), dict_count (t->triangled));
  fprintf (stream, "*NODES\n");
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    fprintf (stream, "%d %g %g\n", n->id, n->x, n->y);
  }
  fprintf (stream, "*TR3N\n");
  dict_reset_iterator (di);
  while (tri = tri_next_triangle (t, di)) {
    fprintf (stream, "%d %d %d %d %d %f %d %s\n",
             tri->id, tri->nodes[0]->id, tri->nodes[1]->id, tri->nodes[2]->id,
             1, 1.0, 1, tri->group->name);
  }
  dict_del_iterator (di);
  return TRUE;
}


